@@ -56,8 +56,16 @@
  ****************************************************************************/
 
 static void
-EventDeviceOnHook(LocalDevicePtr local, SynapticsSHM *para)
+EventDeviceOnHook(LocalDevicePtr local, SynapticsParameters *para)
 {
+    SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
+    BOOL *need_grab;
+
+    if (!priv->proto_data)
+        priv->proto_data = calloc(1, sizeof(BOOL));
+
+    need_grab = (BOOL*)priv->proto_data;
+
     if (para->grab_event_device) {
 	/* Try to grab the event device so that data don't leak to /dev/input/mice */
 	int ret;
@@ -67,52 +75,95 @@ EventDeviceOnHook(LocalDevicePtr local, SynapticsSHM *para)
 		    local->name, errno);
 	}
     }
-}
 
-static void
-EventDeviceOffHook(LocalDevicePtr local)
-{
+    *need_grab = FALSE;
 }
 
 static Bool
-event_query_is_touchpad(int fd)
+event_query_is_touchpad(int fd, BOOL grab)
 {
-    int ret;
-    unsigned long evbits[NBITS(EV_MAX)];
-    unsigned long absbits[NBITS(ABS_MAX)];
-    unsigned long keybits[NBITS(KEY_MAX)];
+    int ret = FALSE, rc;
+    unsigned long evbits[NBITS(EV_MAX)] = {0};
+    unsigned long absbits[NBITS(ABS_MAX)] = {0};
+    unsigned long keybits[NBITS(KEY_MAX)] = {0};
+
+    if (grab)
+    {
+        SYSCALL(rc = ioctl(fd, EVIOCGRAB, (pointer)1));
+        if (rc < 0)
+            return FALSE;
+    }
 
     /* Check for ABS_X, ABS_Y, ABS_PRESSURE and BTN_TOOL_FINGER */
 
-    SYSCALL(ret = ioctl(fd, EVIOCGBIT(0, sizeof(evbits)), evbits));
-    if (ret < 0)
-	return FALSE;
+    SYSCALL(rc = ioctl(fd, EVIOCGBIT(0, sizeof(evbits)), evbits));
+    if (rc < 0)
+	goto unwind;
     if (!TEST_BIT(EV_SYN, evbits) ||
 	!TEST_BIT(EV_ABS, evbits) ||
 	!TEST_BIT(EV_KEY, evbits))
-	return FALSE;
+	goto unwind;
 
-    SYSCALL(ret = ioctl(fd, EVIOCGBIT(EV_ABS, sizeof(absbits)), absbits));
-    if (ret < 0)
-	return FALSE;
+    SYSCALL(rc = ioctl(fd, EVIOCGBIT(EV_ABS, sizeof(absbits)), absbits));
+    if (rc < 0)
+	goto unwind;
     if (!TEST_BIT(ABS_X, absbits) ||
 	!TEST_BIT(ABS_Y, absbits))
-	return FALSE;
+	goto unwind;
 
-    SYSCALL(ret = ioctl(fd, EVIOCGBIT(EV_KEY, sizeof(keybits)), keybits));
-    if (ret < 0)
-	return FALSE;
+    SYSCALL(rc = ioctl(fd, EVIOCGBIT(EV_KEY, sizeof(keybits)), keybits));
+    if (rc < 0)
+	goto unwind;
 
     /* we expect touchpad either report raw pressure or touches */
     if (!TEST_BIT(ABS_PRESSURE, absbits) && !TEST_BIT(BTN_TOUCH, keybits))
-	return FALSE;
+	goto unwind;
     /* all Synaptics-like touchpad report BTN_TOOL_FINGER */
     if (!TEST_BIT(BTN_TOOL_FINGER, keybits))
-	return FALSE;
+	goto unwind;
     if (TEST_BIT(BTN_TOOL_PEN, keybits))
-	return FALSE;			    /* Don't match wacom tablets */
+	goto unwind;			    /* Don't match wacom tablets */
 
-    return TRUE;
+    ret = TRUE;
+
+unwind:
+    if (grab)
+        SYSCALL(ioctl(fd, EVIOCGRAB, (pointer)0));
+
+    return (ret == TRUE);
+}
+
+typedef struct {
+	short vendor;
+	short product;
+	enum TouchpadModel model;
+} model_lookup_t;
+#define PRODUCT_ANY 0x0000
+
+static model_lookup_t model_lookup_table[] = {
+	{0x0002, 0x0007, MODEL_SYNAPTICS},
+	{0x0002, 0x0008, MODEL_ALPS},
+	{0x05ac, PRODUCT_ANY, MODEL_APPLETOUCH},
+	{0x0, 0x0, 0x0}
+};
+
+static void
+event_query_info(LocalDevicePtr local)
+{
+    SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
+    short id[4];
+    int rc;
+    model_lookup_t *model_lookup;
+
+    SYSCALL(rc = ioctl(local->fd, EVIOCGID, id));
+    if (rc < 0)
+        return;
+
+    for(model_lookup = model_lookup_table; model_lookup->vendor; model_lookup++) {
+        if(model_lookup->vendor == id[ID_VENDOR] &&
+           (model_lookup->product == id[ID_PRODUCT] || model_lookup->product == PRODUCT_ANY))
+            priv->model = model_lookup->model;
+    }
 }
 
 /* Query device for axis ranges */
@@ -120,19 +171,22 @@ static void
 event_query_axis_ranges(LocalDevicePtr local)
 {
     SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
-    struct input_absinfo abs;
-    unsigned long absbits[NBITS(ABS_MAX)];
-    unsigned long keybits[NBITS(KEY_MAX)];
+    struct input_absinfo abs = {0};
+    unsigned long absbits[NBITS(ABS_MAX)] = {0};
+    unsigned long keybits[NBITS(KEY_MAX)] = {0};
     char buf[256];
     int rc;
 
     SYSCALL(rc = ioctl(local->fd, EVIOCGABS(ABS_X), &abs));
     if (rc >= 0)
     {
-	xf86Msg(X_INFO, "%s: x-axis range %d - %d\n", local->name,
+	xf86Msg(X_PROBED, "%s: x-axis range %d - %d\n", local->name,
 		abs.minimum, abs.maximum);
 	priv->minx = abs.minimum;
 	priv->maxx = abs.maximum;
+#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
+	priv->resx = abs.resolution;
+#endif
     } else
 	xf86Msg(X_ERROR, "%s: failed to query axis range (%s)\n", local->name,
 		strerror(errno));
@@ -140,10 +194,13 @@ event_query_axis_ranges(LocalDevicePtr local)
     SYSCALL(rc = ioctl(local->fd, EVIOCGABS(ABS_Y), &abs));
     if (rc >= 0)
     {
-	xf86Msg(X_INFO, "%s: y-axis range %d - %d\n", local->name,
+	xf86Msg(X_PROBED, "%s: y-axis range %d - %d\n", local->name,
 		abs.minimum, abs.maximum);
 	priv->miny = abs.minimum;
 	priv->maxy = abs.maximum;
+#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
+	priv->resy = abs.resolution;
+#endif
     } else
 	xf86Msg(X_ERROR, "%s: failed to query axis range (%s)\n", local->name,
 		strerror(errno));
@@ -161,7 +218,7 @@ event_query_axis_ranges(LocalDevicePtr local)
 	SYSCALL(rc = ioctl(local->fd, EVIOCGABS(ABS_PRESSURE), &abs));
 	if (rc >= 0)
 	{
-	    xf86Msg(X_INFO, "%s: pressure range %d - %d\n", local->name,
+	    xf86Msg(X_PROBED, "%s: pressure range %d - %d\n", local->name,
 		    abs.minimum, abs.maximum);
 	    priv->minp = abs.minimum;
 	    priv->maxp = abs.maximum;
@@ -174,7 +231,7 @@ event_query_axis_ranges(LocalDevicePtr local)
     SYSCALL(rc = ioctl(local->fd, EVIOCGABS(ABS_TOOL_WIDTH), &abs));
     if (rc >= 0)
     {
-	xf86Msg(X_INFO, "%s: finger width range %d - %d\n", local->name,
+	xf86Msg(X_PROBED, "%s: finger width range %d - %d\n", local->name,
 		abs.minimum, abs.maximum);
 	priv->minw = abs.minimum;
 	priv->maxw = abs.maximum;
@@ -184,49 +241,66 @@ event_query_axis_ranges(LocalDevicePtr local)
     if (rc >= 0)
     {
 	buf[0] = 0;
-	if ((priv->has_left = TEST_BIT(BTN_LEFT, keybits)))
+	if ((priv->has_left = (TEST_BIT(BTN_LEFT, keybits) != 0)))
 	   strcat(buf, " left");
-	if ((priv->has_right = TEST_BIT(BTN_RIGHT, keybits)))
+	if ((priv->has_right = (TEST_BIT(BTN_RIGHT, keybits) != 0)))
 	   strcat(buf, " right");
-	if ((priv->has_middle = TEST_BIT(BTN_MIDDLE, keybits)))
+	if ((priv->has_middle = (TEST_BIT(BTN_MIDDLE, keybits) != 0)))
 	   strcat(buf, " middle");
-	if ((priv->has_double = TEST_BIT(BTN_TOOL_DOUBLETAP, keybits)))
+	if ((priv->has_double = (TEST_BIT(BTN_TOOL_DOUBLETAP, keybits) != 0)))
 	   strcat(buf, " double");
-	if ((priv->has_triple = TEST_BIT(BTN_TOOL_TRIPLETAP, keybits)))
+	if ((priv->has_triple = (TEST_BIT(BTN_TOOL_TRIPLETAP, keybits) != 0)))
 	   strcat(buf, " triple");
-	xf86Msg(X_INFO, "%s: buttons:%s\n", local->name, buf);
+
+	if ((TEST_BIT(BTN_0, keybits) != 0) ||
+	    (TEST_BIT(BTN_1, keybits) != 0) ||
+	    (TEST_BIT(BTN_2, keybits) != 0) ||
+	    (TEST_BIT(BTN_3, keybits) != 0))
+	{
+	    priv->has_scrollbuttons = 1;
+	    strcat(buf, " scroll-buttons");
+	}
+
+	xf86Msg(X_PROBED, "%s: buttons:%s\n", local->name, buf);
     }
 }
 
 static Bool
-EventQueryHardware(LocalDevicePtr local, struct SynapticsHwInfo *synhw)
+EventQueryHardware(LocalDevicePtr local)
 {
-    if (!event_query_is_touchpad(local->fd))
+    SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
+    BOOL *need_grab = (BOOL*)priv->proto_data;
+
+    if (!event_query_is_touchpad(local->fd, (need_grab) ? *need_grab : TRUE))
 	return FALSE;
 
-    xf86Msg(X_PROBED, "%s touchpad found\n", local->name);
+    xf86Msg(X_PROBED, "%s: touchpad found\n", local->name);
 
     return TRUE;
 }
 
 static Bool
-SynapticsReadEvent(struct CommData *comm, struct input_event *ev)
+SynapticsReadEvent(LocalDevicePtr local, struct input_event *ev)
 {
-    int i, c;
-    unsigned char *pBuf, u;
+    int rc = TRUE;
+    ssize_t len;
 
-    for (i = 0; i < sizeof(struct input_event); i++) {
-	if ((c = XisbRead(comm->buffer)) < 0)
-	    return FALSE;
-	u = (unsigned char)c;
-	pBuf = (unsigned char *)ev;
-	pBuf[i] = u;
+    len = read(local->fd, ev, sizeof(*ev));
+    if (len <= 0)
+    {
+        /* We use X_NONE here because it doesn't alloc */
+        if (errno != EAGAIN)
+            xf86MsgVerb(X_NONE, 0, "%s: Read error %s\n", local->name, strerror(errno));
+        rc = FALSE;
+    } else if (len % sizeof(*ev)) {
+        xf86MsgVerb(X_NONE, 0, "%s: Read error, invalid number of bytes.", local->name);
+        rc = FALSE;
     }
-    return TRUE;
+    return rc;
 }
 
 static Bool
-EventReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
+EventReadHwState(LocalDevicePtr local,
 		 struct SynapticsProtocolOperations *proto_ops,
 		 struct CommData *comm, struct SynapticsHwState *hwRet)
 {
@@ -234,9 +308,9 @@ EventReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
     Bool v;
     struct SynapticsHwState *hw = &(comm->hwState);
     SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
-    SynapticsSHM *para = priv->synpara;
+    SynapticsParameters *para = &priv->synpara;
 
-    while (SynapticsReadEvent(comm, &ev)) {
+    while (SynapticsReadEvent(local, &ev)) {
 	switch (ev.type) {
 	case EV_SYN:
 	    switch (ev.code) {
@@ -250,7 +324,6 @@ EventReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 		else
 		    hw->numFingers = 0;
 		*hwRet = *hw;
-		hw->guest_dx = hw->guest_dy = 0;
 		return TRUE;
 	    }
 	case EV_KEY:
@@ -304,12 +377,6 @@ EventReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 	    case BTN_TOOL_TRIPLETAP:
 		comm->threeFingers = v;
 		break;
-	    case BTN_A:
-		hw->guest_left = v;
-		break;
-	    case BTN_B:
-		hw->guest_right = v;
-		break;
 	    case BTN_TOUCH:
 		if (!priv->has_pressure)
 			hw->z = v ? para->finger_high + 1 : 0;
@@ -332,16 +399,6 @@ EventReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 		break;
 	    }
 	    break;
-	case EV_REL:
-	    switch (ev.code) {
-	    case REL_X:
-		hw->guest_dx = ev.value;
-		break;
-	    case REL_Y:
-		hw->guest_dy = ev.value;
-		break;
-	    }
-	    break;
 	}
     }
     return FALSE;
@@ -358,8 +415,12 @@ static int EventDevOnly(const struct dirent *dir) {
 static void
 EventReadDevDimensions(LocalDevicePtr local)
 {
-    if (event_query_is_touchpad(local->fd))
+    SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
+    BOOL *need_grab = (BOOL*)priv->proto_data;
+
+    if (event_query_is_touchpad(local->fd, (need_grab) ? *need_grab : TRUE))
 	event_query_axis_ranges(local);
+    event_query_info(local);
 }
 
 static Bool
@@ -373,11 +434,11 @@ EventAutoDevProbe(LocalDevicePtr local)
 
     i = scandir(DEV_INPUT_EVENT, &namelist, EventDevOnly, alphasort);
     if (i < 0) {
-		ErrorF("Couldn't open %s\n", DEV_INPUT_EVENT);
+		xf86Msg(X_ERROR, "Couldn't open %s\n", DEV_INPUT_EVENT);
 		return FALSE;
     }
     else if (i == 0) {
-		ErrorF("%s The /dev/input/event* device nodes seem to be missing\n",
+		xf86Msg(X_ERROR, "%s The /dev/input/event* device nodes seem to be missing\n",
 				local->name);
 		free(namelist);
 		return FALSE;
@@ -393,7 +454,7 @@ EventAutoDevProbe(LocalDevicePtr local)
 			if (fd < 0)
 				continue;
 
-			if (event_query_is_touchpad(fd)) {
+			if (event_query_is_touchpad(fd, TRUE)) {
 				touchpad_found = TRUE;
 			    xf86Msg(X_PROBED, "%s auto-dev sets device to %s\n",
 				    local->name, fname);
@@ -407,7 +468,7 @@ EventAutoDevProbe(LocalDevicePtr local)
 	free(namelist);
 
 	if (!touchpad_found) {
-		ErrorF("%s no synaptics event device found\n", local->name);
+		xf86Msg(X_ERROR, "%s no synaptics event device found\n", local->name);
 		return FALSE;
 	}
     return TRUE;
@@ -415,7 +476,7 @@ EventAutoDevProbe(LocalDevicePtr local)
 
 struct SynapticsProtocolOperations event_proto_operations = {
     EventDeviceOnHook,
-    EventDeviceOffHook,
+    NULL,
     EventQueryHardware,
     EventReadHwState,
     EventAutoDevProbe,
