@@ -40,9 +40,21 @@
 #include "ps2comm.h"
 #include "synproto.h"
 #include "synaptics.h"
+#include "synapticsstr.h"
 #include <xf86.h>
 
 #define MAX_UNSYNC_PACKETS 10				/* i.e. 10 to 60 bytes */
+/*
+ * The x/y limits are taken from the Synaptics TouchPad interfacing Guide,
+ * section 2.3.2, which says that they should be valid regardless of the
+ * actual size of the sensor.
+ */
+#define XMIN_NOMINAL 1472
+#define XMAX_NOMINAL 5472
+#define YMIN_NOMINAL 1408
+#define YMAX_NOMINAL 4448
+
+#define XMAX_VALID 6143
 
 /* synaptics queries */
 #define SYN_QUE_IDENTIFY		0x00
@@ -64,17 +76,18 @@
 #define PS2_RES_RESOLUTION(r)	(((r) >> 8) & 0x03)
 #define PS2_RES_SAMPLE_RATE(r)	((r) & 0xff)
 
-/* #define DEBUG */
-
 #ifdef DEBUG
 #define PS2DBG(x) (x)
 #else
 #define PS2DBG(x)
 #endif
 
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 1
-#define DBG(a,b)
-#endif
+struct SynapticsHwInfo {
+    unsigned int model_id;		    /* Model-ID */
+    unsigned int capabilities;		    /* Capabilities */
+    unsigned int ext_cap;		    /* Extended Capabilities */
+    unsigned int identity;		    /* Identification */
+};
 
 /*****************************************************************************
  *	PS/2 Utility functions.
@@ -159,61 +172,6 @@ ps2_send_cmd(int fd, byte c)
 	    ps2_putbyte(fd, PS2_CMD_STATUS_REQUEST));
 }
 
-/*****************************************************************************
- *	Synaptics passthrough functions
- ****************************************************************************/
-
-static Bool
-ps2_getbyte_passthrough(int fd, byte *response)
-{
-    byte ack;
-    int timeout_count;
-#define MAX_RETRY_COUNT 30
-
-    /* Getting a response back through the passthrough could take some time.
-     * Spin a little for the first byte */
-    for (timeout_count = 0;
-	 !ps2_getbyte(fd, &ack) && (timeout_count <= MAX_RETRY_COUNT);
-	 timeout_count++)
-	;
-    /* Do some sanity checking */
-    if ((ack & 0xfc) != 0x84) {
-	PS2DBG(ErrorF("ps2_getbyte_passthrough: expected 0x84 and got: %02x\n",
-		      ack & 0xfc));
-	return FALSE;
-    }
-
-    ps2_getbyte(fd, response);
-    ps2_getbyte(fd, &ack);
-    ps2_getbyte(fd, &ack);
-    if ((ack & 0xcc) != 0xc4) {
-	PS2DBG(ErrorF("ps2_getbyte_passthrough: expected 0xc4 and got: %02x\n",
-		      ack & 0xcc));
-	return FALSE;
-    }
-    ps2_getbyte(fd, &ack);
-    ps2_getbyte(fd, &ack);
-
-    return TRUE;
-}
-
-static Bool
-ps2_putbyte_passthrough(int fd, byte c)
-{
-    byte ack;
-
-    ps2_special_cmd(fd, c);
-    ps2_putbyte(fd, 0xF3);
-    ps2_putbyte(fd, 0x28);
-
-    ps2_getbyte_passthrough(fd, &ack);
-    if (ack != PS2_ACK) {
-	PS2DBG(ErrorF("ps2_putbyte_passthrough: wrong acknowledge 0x%02x\n", ack));
-	return FALSE;
-    }
-    return TRUE;
-}
-
 /*****************************************************************************
  *	Synaptics communications functions
  ****************************************************************************/
@@ -258,30 +216,6 @@ ps2_synaptics_reset(int fd)
     return FALSE;
 }
 
-static Bool
-ps2_synaptics_reset_passthrough(int fd)
-{
-    byte ack;
-
-    /* send reset */
-    ps2_putbyte_passthrough(fd, 0xff);
-    ps2_getbyte_passthrough(fd, &ack);
-    if (ack != 0xaa) {
-	PS2DBG(ErrorF("ps2_synaptics_reset_passthrough: ack was %02x not 0xaa\n", ack));
-	return FALSE;
-    }
-    ps2_getbyte_passthrough(fd, &ack);
-    if (ack != 0x00) {
-	PS2DBG(ErrorF("ps2_synaptics_reset_passthrough: ack was %02x not 0x00\n", ack));
-	return FALSE;
-    }
-
-    /* set defaults, turn on streaming, and enable the mouse */
-    return (ps2_putbyte_passthrough(fd, 0xf6) &&
-	    ps2_putbyte_passthrough(fd, 0xea) &&
-	    ps2_putbyte_passthrough(fd, 0xf4));
-}
-
 /*
  * Read the model-id bytes from the touchpad
  * see also SYN_MODEL_* macros
@@ -299,7 +233,7 @@ ps2_synaptics_model_id(int fd, struct SynapticsHwInfo *synhw)
 	ps2_getbyte(fd, &mi[1]) &&
 	ps2_getbyte(fd, &mi[2])) {
 	synhw->model_id = (mi[0] << 16) | (mi[1] << 8) | mi[2];
-	PS2DBG(ErrorF("mode-id %06X\n", synhw->model_id));
+	PS2DBG(ErrorF("model-id %06X\n", synhw->model_id));
 	PS2DBG(ErrorF("...done.\n"));
 	return TRUE;
     }
@@ -326,8 +260,8 @@ ps2_synaptics_capability(int fd, struct SynapticsHwInfo *synhw)
 	ps2_getbyte(fd, &cap[2])) {
 	synhw->capabilities = (cap[0] << 16) | (cap[1] << 8) | cap[2];
 	PS2DBG(ErrorF("capabilities %06X\n", synhw->capabilities));
-	if (SYN_CAP_VALID(*synhw)) {
-	    if (SYN_EXT_CAP_REQUESTS(*synhw)) {
+	if (SYN_CAP_VALID(synhw)) {
+	    if (SYN_EXT_CAP_REQUESTS(synhw)) {
 		if (ps2_send_cmd(fd, SYN_QUE_EXT_CAPAB) &&
 		    ps2_getbyte(fd, &cap[0]) &&
 		    ps2_getbyte(fd, &cap[1]) &&
@@ -365,7 +299,7 @@ ps2_synaptics_identify(int fd, struct SynapticsHwInfo *synhw)
 	ps2_getbyte(fd, &id[2])) {
 	synhw->identity = (id[0] << 16) | (id[1] << 8) | id[2];
 	PS2DBG(ErrorF("ident %06X\n", synhw->identity));
-	if (SYN_ID_IS_SYNAPTICS(*synhw)) {
+	if (SYN_ID_IS_SYNAPTICS(synhw)) {
 	    PS2DBG(ErrorF("...done.\n"));
 	    return TRUE;
 	}
@@ -388,9 +322,8 @@ ps2_synaptics_disable_device(int fd)
 }
 
 static Bool
-ps2_query_is_synaptics(int fd)
+ps2_query_is_synaptics(int fd, struct SynapticsHwInfo* synhw)
 {
-    struct SynapticsHwInfo synhw;
     int i;
 
     for (i = 0; i < 3; i++) {
@@ -400,10 +333,10 @@ ps2_query_is_synaptics(int fd)
 
     xf86WaitForInput(fd, 20000);
     xf86FlushInput(fd);
-    if (ps2_synaptics_identify(fd, &synhw)) {
+    if (ps2_synaptics_identify(fd, synhw)) {
 	return TRUE;
     } else {
-	ErrorF("Query no Synaptics: %06X\n", synhw.identity);
+	xf86Msg(X_ERROR, "Query no Synaptics: %06X\n", synhw->identity);
 	return FALSE;
     }
 }
@@ -411,44 +344,38 @@ ps2_query_is_synaptics(int fd)
 void
 ps2_print_ident(const struct SynapticsHwInfo *synhw)
 {
-    xf86Msg(X_PROBED, " Synaptics Touchpad, model: %d\n", SYN_ID_MODEL(*synhw));
-    xf86Msg(X_PROBED, " Firmware: %d.%d\n", SYN_ID_MAJOR(*synhw),
-	    SYN_ID_MINOR(*synhw));
+    xf86Msg(X_PROBED, " Synaptics Touchpad, model: %d\n", SYN_ID_MODEL(synhw));
+    xf86Msg(X_PROBED, " Firmware: %d.%d\n", SYN_ID_MAJOR(synhw),
+	    SYN_ID_MINOR(synhw));
 
-    if (SYN_MODEL_ROT180(*synhw))
+    if (SYN_MODEL_ROT180(synhw))
 	xf86Msg(X_PROBED, " 180 degree mounted touchpad\n");
-    if (SYN_MODEL_PORTRAIT(*synhw))
+    if (SYN_MODEL_PORTRAIT(synhw))
 	xf86Msg(X_PROBED, " portrait touchpad\n");
-    xf86Msg(X_PROBED, " Sensor: %d\n", SYN_MODEL_SENSOR(*synhw));
-    if (SYN_MODEL_NEWABS(*synhw))
+    xf86Msg(X_PROBED, " Sensor: %d\n", SYN_MODEL_SENSOR(synhw));
+    if (SYN_MODEL_NEWABS(synhw))
 	xf86Msg(X_PROBED, " new absolute packet format\n");
-    if (SYN_MODEL_PEN(*synhw))
+    if (SYN_MODEL_PEN(synhw))
 	xf86Msg(X_PROBED, " pen detection\n");
 
-    if (SYN_CAP_EXTENDED(*synhw)) {
+    if (SYN_CAP_EXTENDED(synhw)) {
 	xf86Msg(X_PROBED, " Touchpad has extended capability bits\n");
-	if (SYN_CAP_MULTI_BUTTON_NO(*synhw))
+	if (SYN_CAP_MULTI_BUTTON_NO(synhw))
 	    xf86Msg(X_PROBED, " -> %d multi buttons, i.e. besides standard buttons\n",
-		    (int)(SYN_CAP_MULTI_BUTTON_NO(*synhw)));
-	if (SYN_CAP_MIDDLE_BUTTON(*synhw))
+		    (int)(SYN_CAP_MULTI_BUTTON_NO(synhw)));
+	if (SYN_CAP_MIDDLE_BUTTON(synhw))
 	    xf86Msg(X_PROBED, " -> middle button\n");
-	if (SYN_CAP_FOUR_BUTTON(*synhw))
+	if (SYN_CAP_FOUR_BUTTON(synhw))
 	    xf86Msg(X_PROBED, " -> four buttons\n");
-	if (SYN_CAP_MULTIFINGER(*synhw))
+	if (SYN_CAP_MULTIFINGER(synhw))
 	    xf86Msg(X_PROBED, " -> multifinger detection\n");
-	if (SYN_CAP_PALMDETECT(*synhw))
+	if (SYN_CAP_PALMDETECT(synhw))
 	    xf86Msg(X_PROBED, " -> palm detection\n");
-	if (SYN_CAP_PASSTHROUGH(*synhw))
+	if (SYN_CAP_PASSTHROUGH(synhw))
 	    xf86Msg(X_PROBED, " -> pass-through port\n");
     }
 }
 
-
-static void
-PS2DeviceOnHook(LocalDevicePtr local, SynapticsSHM* para)
-{
-}
-
 static void
 PS2DeviceOffHook(LocalDevicePtr local)
 {
@@ -457,12 +384,18 @@ PS2DeviceOffHook(LocalDevicePtr local)
 }
 
 static Bool
-PS2QueryHardware(LocalDevicePtr local, struct SynapticsHwInfo *synhw)
+PS2QueryHardware(LocalDevicePtr local)
 {
     int mode;
+    SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
+    struct SynapticsHwInfo *synhw;
+
+    if (!priv->proto_data)
+        priv->proto_data = calloc(1, sizeof(struct SynapticsHwInfo));
+    synhw = (struct SynapticsHwInfo*)priv->proto_data;
 
     /* is the synaptics touchpad active? */
-    if (!ps2_query_is_synaptics(local->fd))
+    if (!ps2_query_is_synaptics(local->fd, synhw))
 	return FALSE;
 
     xf86Msg(X_PROBED, "%s synaptics touchpad found\n", local->name);
@@ -480,29 +413,13 @@ PS2QueryHardware(LocalDevicePtr local, struct SynapticsHwInfo *synhw)
 	return FALSE;
 
     mode = SYN_BIT_ABSOLUTE_MODE | SYN_BIT_HIGH_RATE;
-    if (SYN_ID_MAJOR(*synhw) >= 4)
+    if (SYN_ID_MAJOR(synhw) >= 4)
 	mode |= SYN_BIT_DISABLE_GESTURE;
-    if (SYN_CAP_EXTENDED(*synhw))
+    if (SYN_CAP_EXTENDED(synhw))
 	mode |= SYN_BIT_W_MODE;
     if (!ps2_synaptics_set_mode(local->fd, mode))
 	return FALSE;
 
-    /* Check to see if the host mouse supports a guest */
-    synhw->hasGuest = FALSE;
-    if (SYN_CAP_PASSTHROUGH(*synhw)) {
-        synhw->hasGuest = TRUE;
-
-	/* Enable the guest mouse.  Set it to relative mode, three byte
-	 * packets */
-
-	/* Disable the host to talk to the guest */
-	ps2_synaptics_disable_device(local->fd);
-	/* Reset it, set defaults, streaming and enable it */
-	if (!ps2_synaptics_reset_passthrough(local->fd)) {
-	    synhw->hasGuest = FALSE;
-	}
-    }
-
     ps2_synaptics_enable_device(local->fd);
 
     ps2_print_ident(synhw);
@@ -517,25 +434,25 @@ static Bool
 ps2_packet_ok(struct SynapticsHwInfo *synhw, struct CommData *comm)
 {
     unsigned char *buf = comm->protoBuf;
-    int newabs = SYN_MODEL_NEWABS(*synhw);
+    int newabs = SYN_MODEL_NEWABS(synhw);
 
     if (newabs ? ((buf[0] & 0xC0) != 0x80) : ((buf[0] & 0xC0) != 0xC0)) {
-	DBG(4, ErrorF("Synaptics driver lost sync at 1st byte\n"));
+	DBG(4, "Synaptics driver lost sync at 1st byte\n");
 	return FALSE;
     }
 
     if (!newabs && ((buf[1] & 0x60) != 0x00)) {
-	DBG(4, ErrorF("Synaptics driver lost sync at 2nd byte\n"));
+	DBG(4, "Synaptics driver lost sync at 2nd byte\n");
 	return FALSE;
     }
 
     if ((newabs ? ((buf[3] & 0xC0) != 0xC0) : ((buf[3] & 0xC0) != 0x80))) {
-	DBG(4, ErrorF("Synaptics driver lost sync at 4th byte\n"));
+	DBG(4, "Synaptics driver lost sync at 4th byte\n");
 	return FALSE;
     }
 
     if (!newabs && ((buf[4] & 0x60) != 0x00)) {
-	DBG(4, ErrorF("Synaptics driver lost sync at 5th byte\n"));
+	DBG(4, "Synaptics driver lost sync at 5th byte\n");
 	return FALSE;
     }
 
@@ -557,16 +474,16 @@ ps2_synaptics_get_packet(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 	/* test if there is a reset sequence received */
 	if ((c == 0x00) && (comm->lastByte == 0xAA)) {
 	    if (xf86WaitForInput(local->fd, 50000) == 0) {
-		DBG(7, ErrorF("Reset received\n"));
-		proto_ops->QueryHardware(local, synhw);
+		DBG(7, "Reset received\n");
+		proto_ops->QueryHardware(local);
 	    } else
-		DBG(3, ErrorF("faked reset received\n"));
+		DBG(3, "faked reset received\n");
 	}
 	comm->lastByte = u;
 
 	/* to avoid endless loops */
 	if (count++ > 30) {
-	    ErrorF("Synaptics driver lost sync... got gigantic packet!\n");
+	    xf86Msg(X_ERROR, "Synaptics driver lost sync... got gigantic packet!\n");
 	    return FALSE;
 	}
 
@@ -583,8 +500,8 @@ ps2_synaptics_get_packet(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 		comm->outOfSync++;
 		if (comm->outOfSync > MAX_UNSYNC_PACKETS) {
 		    comm->outOfSync = 0;
-		    DBG(3, ErrorF("Synaptics synchronization lost too long -> reset touchpad.\n"));
-		    proto_ops->QueryHardware(local, synhw); /* including a reset */
+		    DBG(3, "Synaptics synchronization lost too long -> reset touchpad.\n");
+		    proto_ops->QueryHardware(local); /* including a reset */
 		    continue;
 		}
 	    }
@@ -593,7 +510,7 @@ ps2_synaptics_get_packet(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 	if (comm->protoBufTail >= 6) { /* Full packet received */
 	    if (comm->outOfSync > 0) {
 		comm->outOfSync = 0;
-		DBG(4, ErrorF("Synaptics driver resynced.\n"));
+		DBG(4, "Synaptics driver resynced.\n");
 	    }
 	    comm->protoBufTail = 0;
 	    return TRUE;
@@ -604,37 +521,32 @@ ps2_synaptics_get_packet(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 }
 
 static Bool
-PS2ReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
+PS2ReadHwState(LocalDevicePtr local,
 	       struct SynapticsProtocolOperations *proto_ops,
 	       struct CommData *comm, struct SynapticsHwState *hwRet)
 {
-    int newabs = SYN_MODEL_NEWABS(*synhw);
     unsigned char *buf = comm->protoBuf;
     struct SynapticsHwState *hw = &(comm->hwState);
+    SynapticsPrivate *priv = (SynapticsPrivate *)local->private;
+    SynapticsParameters *para = &priv->synpara;
+    struct SynapticsHwInfo *synhw;
+    int newabs;
     int w, i;
 
+    synhw = (struct SynapticsHwInfo*)priv->proto_data;
+    if (!synhw)
+    {
+        xf86Msg(X_ERROR,
+                "%s: PS2ReadHwState, synhw is NULL. This is a bug.\n",
+                local->name);
+        return FALSE;
+    }
+
+    newabs = SYN_MODEL_NEWABS(synhw);
+
     if (!ps2_synaptics_get_packet(local, synhw, proto_ops, comm))
 	return FALSE;
 
-    /* Handle guest packets */
-    hw->guest_dx = hw->guest_dy = 0;
-    if (newabs && synhw->hasGuest) {
-	w = (((buf[0] & 0x30) >> 2) |
-	     ((buf[0] & 0x04) >> 1) |
-	     ((buf[3] & 0x04) >> 2));
-	if (w == 3) {	       /* If w is 3, this is a guest packet */
-	    if (buf[4] != 0)
-		hw->guest_dx =   buf[4] - ((buf[1] & 0x10) ? 256 : 0);
-	    if (buf[5] != 0)
-		hw->guest_dy = -(buf[5] - ((buf[1] & 0x20) ? 256 : 0));
-	    hw->guest_left  = (buf[1] & 0x01) ? TRUE : FALSE;
-	    hw->guest_mid   = (buf[1] & 0x04) ? TRUE : FALSE;
-	    hw->guest_right = (buf[1] & 0x02) ? TRUE : FALSE;
-	    *hwRet = *hw;
-	    return TRUE;
-	}
-    }
-
     /* Handle normal packets */
     hw->x = hw->y = hw->z = hw->numFingers = hw->fingerWidth = 0;
     hw->left = hw->right = hw->up = hw->down = hw->middle = FALSE;
@@ -642,7 +554,7 @@ PS2ReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 	hw->multi[i] = FALSE;
 
     if (newabs) {			    /* newer protos...*/
-	DBG(7, ErrorF("using new protocols\n"));
+	DBG(7, "using new protocols\n");
 	hw->x = (((buf[3] & 0x10) << 8) |
 		 ((buf[1] & 0x0f) << 8) |
 		 buf[4]);
@@ -658,11 +570,11 @@ PS2ReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 	hw->left  = (buf[0] & 0x01) ? 1 : 0;
 	hw->right = (buf[0] & 0x02) ? 1 : 0;
 
-	if (SYN_CAP_EXTENDED(*synhw)) {
-	    if (SYN_CAP_MIDDLE_BUTTON(*synhw)) {
+	if (SYN_CAP_EXTENDED(synhw)) {
+	    if (SYN_CAP_MIDDLE_BUTTON(synhw)) {
 		hw->middle = ((buf[0] ^ buf[3]) & 0x01) ? 1 : 0;
 	    }
-	    if (SYN_CAP_FOUR_BUTTON(*synhw)) {
+	    if (SYN_CAP_FOUR_BUTTON(synhw)) {
 		hw->up = ((buf[3] & 0x01)) ? 1 : 0;
 		if (hw->left)
 		    hw->up = !hw->up;
@@ -670,9 +582,9 @@ PS2ReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 		if (hw->right)
 		    hw->down = !hw->down;
 	    }
-	    if (SYN_CAP_MULTI_BUTTON_NO(*synhw)) {
+	    if (SYN_CAP_MULTI_BUTTON_NO(synhw)) {
 		if ((buf[3] & 2) ? !hw->right : hw->right) {
-		    switch (SYN_CAP_MULTI_BUTTON_NO(*synhw) & ~0x01) {
+		    switch (SYN_CAP_MULTI_BUTTON_NO(synhw) & ~0x01) {
 		    default:
 			break;
 		    case 8:
@@ -692,7 +604,7 @@ PS2ReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 	    }
 	}
     } else {			    /* old proto...*/
-	DBG(7, ErrorF("using old protocol\n"));
+	DBG(7, "using old protocol\n");
 	hw->x = (((buf[1] & 0x1F) << 8) |
 		 buf[2]);
 	hw->y = (((buf[4] & 0x1F) << 8) |
@@ -709,20 +621,20 @@ PS2ReadHwState(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
 
     hw->y = YMAX_NOMINAL + YMIN_NOMINAL - hw->y;
 
-    if (hw->z > 0) {
+    if (hw->z >= para->finger_high) {
 	int w_ok = 0;
 	/*
 	 * Use capability bits to decide if the w value is valid.
 	 * If not, set it to 5, which corresponds to a finger of
 	 * normal width.
 	 */
-	if (SYN_CAP_EXTENDED(*synhw)) {
+	if (SYN_CAP_EXTENDED(synhw)) {
 	    if ((w >= 0) && (w <= 1)) {
-		w_ok = SYN_CAP_MULTIFINGER(*synhw);
+		w_ok = SYN_CAP_MULTIFINGER(synhw);
 	    } else if (w == 2) {
-		w_ok = SYN_MODEL_PEN(*synhw);
+		w_ok = SYN_MODEL_PEN(synhw);
 	    } else if ((w >= 4) && (w <= 15)) {
-		w_ok = SYN_CAP_PALMDETECT(*synhw);
+		w_ok = SYN_CAP_PALMDETECT(synhw);
 	    }
 	}
 	if (!w_ok)
@@ -755,10 +667,10 @@ PS2AutoDevProbe(LocalDevicePtr local)
 }
 
 struct SynapticsProtocolOperations psaux_proto_operations = {
-    PS2DeviceOnHook,
+    NULL,
     PS2DeviceOffHook,
     PS2QueryHardware,
     PS2ReadHwState,
     PS2AutoDevProbe,
-    NULL /* ReadDevDimensions */
+    SynapticsDefaultDimensions
 };
