@@ -50,12 +50,6 @@ struct SynapticsHwState {
 
     Bool multi[8];
     Bool middle;		/* Some ALPS touchpads have a middle button */
-
-    Bool guest_left;		/* guest device */
-    Bool guest_mid;
-    Bool guest_right;
-    int  guest_dx;
-    int  guest_dy;
 };
 
 struct CommData {
@@ -84,18 +78,19 @@ enum SynapticsProtocol {
     SYN_PROTO_ALPS		/* ALPS touchpad protocol */
 };
 
-struct _SynapticsSHM;
+struct _SynapticsParameters;
 struct SynapticsHwInfo;
 struct CommData;
 
 struct SynapticsProtocolOperations {
-    void (*DeviceOnHook)(LocalDevicePtr local, struct _SynapticsSHM *para);
+    void (*DeviceOnHook)(LocalDevicePtr local, struct _SynapticsParameters *para);
     void (*DeviceOffHook)(LocalDevicePtr local);
-    Bool (*QueryHardware)(LocalDevicePtr local, struct SynapticsHwInfo *synhw);
-    Bool (*ReadHwState)(LocalDevicePtr local, struct SynapticsHwInfo *synhw,
+    Bool (*QueryHardware)(LocalDevicePtr local);
+    Bool (*ReadHwState)(LocalDevicePtr local,
 			struct SynapticsProtocolOperations *proto_ops,
 			struct CommData *comm, struct SynapticsHwState *hwRet);
     Bool (*AutoDevProbe)(LocalDevicePtr local);
+    void (*ReadDevDimensions)(LocalDevicePtr local);
 };
 
 extern struct SynapticsProtocolOperations psaux_proto_operations;
